from flask_discord_interactions.models.channel import Channel, ChannelType
from flask_discord_interactions.models.user import User, Member
from flask_discord_interactions.models.message import Message, ResponseType
from flask_discord_interactions.models.embed import Embed
from flask_discord_interactions.models.component import (
    ComponentType,
    ActionRow,
    Button,
    SelectMenu,
    SelectMenuOption,
    Component,
    ButtonStyles,
    TextInput,
    TextStyles,
)
from flask_discord_interactions.models.modal import Modal
from flask_discord_interactions.models.command import ApplicationCommandType
from flask_discord_interactions.models.permission import Permission
from flask_discord_interactions.models.role import Role
from flask_discord_interactions.models.utils import LoadableDataclass
from flask_discord_interactions.models.autocomplete import (
    Autocomplete,
    AutocompleteResult,
)
from flask_discord_interactions.models.option import Option, CommandOptionType

__all__ = [
    "Channel",
    "ChannelType",
    "User",
    "Member",
    "Message",
    "ResponseType",
    "Embed",
    "ComponentType",
    "ActionRow",
    "Button",
    "SelectMenu",
    "SelectMenuOption",
    "Component",
    "ButtonStyles",
    "TextInput",
    "TextStyles",
    "Modal",
    "ApplicationCommandType",
    "CommandOptionType",
    "Permission",
    "Role",
    "LoadableDataclass",
    "Autocomplete",
    "AutocompleteResult",
    "Option",
]



from flask_discord_interactions import (
    Message,
    ResponseType,
    ActionRow,
    Button,
    ComponentType,
    ButtonStyles,
    SelectMenu,
    SelectMenuOption,
)


def test_parse_arguments(discord, client):
    @discord.custom_handler()
    def handler(ctx, string_arg, int_arg: int):
        return f"String: {string_arg}, type(int_arg): {type(int_arg)}"

    assert (
        client.run_handler(handler, "hello!", "42").content
        == "String: hello!, type(int_arg): <class 'int'>"
    )


def test_action_row(discord, client):
    @discord.command()
    def action_row(ctx):
        return Message(content="Hi!", components=[ActionRow()])

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [{"type": ComponentType.ACTION_ROW}],
        },
    }

    assert client.run("action_row").dump() == expected


def test_button(discord, client):
    @discord.command()
    def button(ctx):
        return Message(
            content="Hi!",
            components=[
                ActionRow(
                    components=[
                        Button(
                            style=ButtonStyles.PRIMARY,
                            custom_id="my_button",
                            label="My Button",
                        )
                    ]
                )
            ],
        )

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [
                {
                    "type": ComponentType.ACTION_ROW,
                    "components": [
                        {
                            "type": ComponentType.BUTTON,
                            "style": ButtonStyles.PRIMARY,
                            "custom_id": "my_button",
                            "label": "My Button",
                            "disabled": False,
                        }
                    ],
                }
            ],
        },
    }

    assert client.run("button").dump() == expected


def test_select_menu(discord, client):
    @discord.command()
    def selectmenu(ctx):
        return Message(
            content="Hi!",
            components=[
                ActionRow(
                    components=[
                        SelectMenu(
                            custom_id="my_menu",
                            placeholder="Choose an option",
                            options=[
                                SelectMenuOption(label="Option 1", value="option1"),
                                SelectMenuOption(label="Option 2", value="option2"),
                            ],
                        )
                    ]
                )
            ],
        )

    expected = {
        "type": ResponseType.CHANNEL_MESSAGE_WITH_SOURCE,
        "data": {
            "content": "Hi!",
            "embeds": None,
            "flags": 0,
            "tts": False,
            "allowed_mentions": {"parse": ["roles", "users", "everyone"]},
            "components": [
                {
                    "type": ComponentType.ACTION_ROW,
                    "components": [
                        {
                            "type": ComponentType.SELECT_MENU,
                            "custom_id": "my_menu",
                            "placeholder": "Choose an option",
                            "disabled": False,
                            "options": [
                                {
                                    "label": "Option 1",
                                    "value": "option1",
                                    "default": False,
                                },
                                {
                                    "label": "Option 2",
                                    "value": "option2",
                                    "default": False,
                                },
                            ],
                            "max_values": 1,
                            "min_values": 1,
                        }
                    ],
                }
            ],
        },
    }

    assert client.run("selectmenu").dump() == expected



import inspect


class LoadableDataclass:
    @classmethod
    def from_dict(cls, data):
        """
        Construct the dataclass from a dictionary, skipping any keys
        in the dictionary that do not correspond to fields of the class.

        Parameters
        ----------
        data
            A dictionary of fields to set on the dataclass.
        """
        return cls(
            **{k: v for k, v in data.items() if k in inspect.signature(cls).parameters}
        )



class ResponseType:
    "Represents the different response type integers."
    PONG = 1
    CHANNEL_MESSAGE_WITH_SOURCE = 4
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5
    DEFERRED_UPDATE_MESSAGE = 6
    UPDATE_MESSAGE = 7
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8
    MODAL = 9



import asyncio

import pytest
from quart import Quart

from flask_discord_interactions import (
    DiscordInteractions,
    Client,
    Message,
    AsyncContext,
)


@pytest.fixture()
def quart_discord():
    app = Quart(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)
    return discord, Client(discord)


@pytest.mark.asyncio
async def test_async_ping(quart_discord):
    discord, client = quart_discord

    @discord.command()
    async def ping(ctx, pong: str = "pong"):
        "Respond with a friendly 'pong'!"
        return f"Pong {pong}!"

    assert (await client.run("ping")).content == "Pong pong!"


@pytest.mark.asyncio
async def test_await_in_command(quart_discord):
    discord, client = quart_discord

    @discord.command()
    async def wait(ctx):
        await asyncio.sleep(0.01)
        return "Hi!"

    assert (await client.run("wait")).content == "Hi!"


@pytest.mark.asyncio
async def test_mixed_commands(quart_discord):
    discord, client = quart_discord

    @discord.command()
    async def async_command(ctx):
        return "Async"

    @discord.command()
    def not_async_command(ctx):
        return "Not Async"

    assert (await client.run("async_command")).content == "Async"
    assert client.run("not_async_command").content == "Not Async"


@pytest.mark.asyncio
async def test_followup(quart_discord):
    discord, client = quart_discord

    followup_task = None

    @discord.command()
    async def followup(ctx):
        nonlocal followup_task

        async def do_followup():
            await ctx.edit(f"Hello!")

        followup_task = asyncio.create_task(do_followup())
        return Message(deferred=True)

    with client.context(AsyncContext()):
        await client.run("followup")
    await followup_task



from dataclasses import dataclass
from typing import Any

from flask_discord_interactions.models import User, Member, Channel, Role


class CommandOptionType:
    "Represents the different option type integers."
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10


@dataclass
class Option:
    """
    Represents an option provided to a slash command.

    Attributes
    ----------
    name
        The name of the option.
    type
        The type of the option. Provide either a value of
        :class:`.CommandOptionType` or a type (e.g. ``str``).
    description
        The description of the option. Defaults to "No description."
    required
        Whether the option is required. Defaults to ``False``.
    autocomplete
        Whether the option should be autocompleted. Defaults to ``False``.
        Set to ``True`` if you have an autocomplete handler for this command.
    value
        Only present on incoming options passed to autocomplete objects. You
        shouldn't set this yourself. Represents the value that the user is
        currently typing.
    focused
        Only present on incoming options passed to autocomplete objects. True
        if the user is currently typing this option.
    """

    name: str
    type: int

    description: str = "No description"
    required: bool = False
    autocomplete: bool = False

    value: Any = None
    focused: bool = None

    def __post_init__(self):
        if isinstance(self.type, type):
            if self.type == str:
                self.type = CommandOptionType.STRING
            elif self.type == int:
                self.type = CommandOptionType.INTEGER
            elif self.type == bool:
                self.type = CommandOptionType.BOOLEAN
            elif self.type in [User, Member]:
                self.type = CommandOptionType.USER
            elif self.type == Channel:
                self.type = CommandOptionType.CHANNEL
            elif self.type == Role:
                self.type = CommandOptionType.ROLE
            elif self.type == float:
                self.type = CommandOptionType.NUMBER
            else:
                raise ValueError(f"Unknown type {self.type}")

    @classmethod
    def from_data(cls, data):
        "Load this option from incoming Interaction data."
        return cls(
            name=data["name"],
            type=data["type"],
            value=data.get("value"),
            focused=data.get("focused"),
        )

    def dump(self):
        "Return this option as as a dict for registration with Discord."
        data = {
            "name": self.name,
            "type": self.type,
            "description": self.description,
            "required": self.required,
            "autocomplete": self.autocomplete,
        }
        return data



from flask_discord_interactions.tests.conftest import client
from flask_discord_interactions.context import ApplicationCommandType
import pytest


def test_invalid_name(discord):
    with pytest.raises(ValueError):

        @discord.command()
        def this_is_a_very_long_command_name_that_will_raise_an_error(ctx):
            return "this shouldn't work..."


def test_provided_name(discord, client):
    @discord.command(name="other")
    def fn(ctx):
        return "hi"

    assert client.run("other").content == "hi"


def test_uppercase_name(discord):
    with pytest.raises(ValueError):

        @discord.command()
        def UPPERCASE(ctx):
            return "this shouldn't work..."


def test_invalid_character(discord):
    with pytest.raises(ValueError):

        @discord.command(name="invalid&character")
        def invalid_character(ctx):
            return "this shouldn't work..."


def test_type_user_valid(discord, client):
    @discord.command(name="testuser", type=ApplicationCommandType.USER)
    def fn(ctx, user):
        return "hi user"

    assert client.run("testuser").content == "hi user"


def test_type_message_valid(discord, client):
    @discord.command(name="testmessage", type=ApplicationCommandType.MESSAGE)
    def fn(ctx, message):
        return "message"

    assert client.run("testmessage").content == "message"


def test_type_user_valid_uppercase(discord, client):
    @discord.command(name="TEST USER", type=ApplicationCommandType.USER)
    def fn(ctx, user):
        return "hi user"

    assert client.run("TEST USER").content == "hi user"


def test_type_message_valid_uppercase(discord, client):
    @discord.command(name="TEST MSG", type=ApplicationCommandType.MESSAGE)
    def fn(ctx, message):
        return "message"

    assert client.run("TEST MSG").content == "message"



import pytest

from flask import Flask
from flask_discord_interactions import DiscordInteractions, SlashCommand, Response


def test_slash_command(discord, client):
    with pytest.deprecated_call():
        command = SlashCommand(lambda ctx: "ping", "ping", "No description", [], [])

    # make sure the object still works
    assert command.name == "ping"


def test_response():
    with pytest.deprecated_call():
        response = Response("Hello this is my response")

    assert response.content == "Hello this is my response"


def test_add_slash_command(discord):
    with pytest.deprecated_call():
        discord.add_slash_command(lambda ctx: "ping", "ping", "test test test", [], [])

    assert discord.discord_commands["ping"].description == "test test test"


def test_update_slash_commands():
    app = Flask(__name__)
    app.config["DONT_VALIDATE_SIGNATURE"] = True
    app.config["DONT_REGISTER_WITH_DISCORD"] = True

    discord = DiscordInteractions(app)

    with pytest.deprecated_call():
        discord.update_slash_commands()



from typing import Union
from flask_discord_interactions.models.message import ResponseType


class Autocomplete:
    """
    Represents the type of an option that can be autocompleted.

    Attributes
    ----------
    t
        The underlying type of the option.
    """

    def __init__(self, t: type):
        self.t = t


class AutocompleteResult:
    """
    Represents the result of an autocomplete handler.

    Attributes
    ----------
    choices
        A dict mapping the displayed name of each choice to its value passed to
        your command.
    """

    def __init__(self, choices=[]):
        self.choices = choices

    def dump(self):
        "Return this result as a complete interaction response."
        return {
            "type": ResponseType.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT,
            "data": {"choices": self.choices},
        }

    @staticmethod
    def from_return_value(value: Union[dict, list, "AutocompleteResult"]):
        """
        Converts the return value of an autocomplete handler to an
        AutocompleteResult.

        Paramters
        ---------
        value
            The return value of an autocomplete handler.

        Returns
        -------
        AutocompleteResult
            The AutocompleteResult that corresponds to the return value.
        """
        if isinstance(value, AutocompleteResult):
            return value
        elif isinstance(value, dict):
            return AutocompleteResult([value])
        elif isinstance(value, list) and all(
            isinstance(choice, dict) for choice in value
        ):
            return AutocompleteResult(value)
        elif isinstance(value, list):
            return AutocompleteResult(
                [{"name": str(choice), "value": choice} for choice in value]
            )
